/*  cluster.c  */

#ifdef __MSDOS__
#ifndef __COMPACT__
#error Memory model COMPACT required
#endif /* __COMPACT__ */
#include <alloc.h>
#include <dir.h>
#endif /* __MSDOS__ */

#include <ctype.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <values.h>

#define BUFSIZE 2048

struct VEC
{
    char
	*name;
    float
	*v;
    int
	cl,
        nr;
};

struct CLUSTER
{
    float
	distance,
        center;
};

struct VEC
    **vec = NULL,
    **vec_tmp = NULL;

struct CLUSTER
    *cl;

float
    **d;

int
    vec_max = 0,
    vec_n = 0,
    vec_size,
    verbose = 1,
    input_line;

char
    *programname,
    buffer [BUFSIZE],
    buf1 [BUFSIZE],
    buf2 [BUFSIZE],
    buf3 [BUFSIZE],
    *infile = NULL,
    *baseout = NULL,
    *no_mem_buffer,
    Out_of_memory [] = "Out of memory";

void
    get_programname (char const *argv0),
    syntax (void),
    process_args (int argc, char *argv []),
    setup (void),
    read_file (void),
    save_file (void),
    errit (char const *format, ...),
    *m_malloc (size_t size),
    *m_realloc (void *block, size_t size);
char
    *get_arg (int *argc, char *argv [], int *index),
    *m_strdup (const char *s),
    *to_ps (char const *string);
int
    getline (FILE *fp, int required, char const *filename);
float
    square_vecdiff (float *v1, float *v2);
FILE
    *r_fopen (char const *filename),
    *w_fopen (char const *filename);
struct VEC
    *store_vec (char *name);

int main (int argc, char *argv [])
{
    no_mem_buffer = (char *)malloc (1024);

    get_programname (argv [0]);

    process_args (argc, argv);

    if (! infile)
        errit ("Naam input-file ontbreekt");
    if (! baseout)
	baseout = "clus_out";

    read_file ();

    setup ();

    save_file ();

    return 0;
}

void read_file ()
{
    FILE
	*fp;
    struct VEC
	*v;
    int
	i;

    fp = r_fopen (infile);
    getline (fp, 1, infile);
    if (sscanf (buffer, "%i", &vec_size) != 1)
	errit (
	    "file \"%s\", line %i\nVectorsize expected",
	    infile,
	    input_line
	);
    while (getline (fp, 0, NULL)) {
	if (sscanf (buffer, " %s", buf1) != 1)
	    errit ("file \"%s\", line %i", infile, input_line);
	v = store_vec (buf1);
	v->v = (float *) m_malloc (vec_size * sizeof (float));
	for (i = 0; i < vec_size; i++) {
	    getline (fp, 1, infile);
	    if (sscanf (buffer, "%f", &(v->v[i])) != 1)
		errit (
		    "file \"%s\", line %i\n"
		    "Missing value for vector \'%s\'",
		    infile,
		    input_line,
		    v->name
		);
	}
    }
    fclose (fp);
}

void save_file ()
{
    int
	i,
	j,
	k,
	p,
	max_len,
	cl1,
	cl2;
    float
	min,
	maxdiff;
    FILE
	*fp;

    max_len = 0;
    for (i = 0; i < vec_n; i++)
	if ((j = strlen (vec [i]->name)) > max_len)
	    max_len = j;

    if (verbose) {
        printf ("Calculating Sums of Square Differences");
        fflush (stdout);
    }
    for (i = 0; i < vec_n; i++) {
        if (verbose) {
            printf (".");
	    fflush (stdout);
        }
        for (j = i; j < vec_n; j++)
            d [i][j] = d [j][i] = square_vecdiff (vec [i]->v, vec [j]->v);
    }
    if (verbose) {
        printf ("\n");
        fflush (stdout);
    }

    if (verbose) {
	printf ("Sorting");
	fflush (stdout);
    }
    for (i = 0; i < vec_n; i++)
	vec [i]->cl = i;
    for (i = 1; i < vec_n; i++) {
	if (verbose) {
	    printf (".");
	    fflush (stdout);
	}
	min = MAXFLOAT;
	for (j = 0; j < vec_n; j++)
	    for (k = j + 1; k < vec_n; k++)
		if (d [vec[j]->nr][vec[k]->nr] < min && vec[j]->cl != vec[k]->cl) {
		    min = d [vec[j]->nr][vec[k]->nr];
		    cl1 = vec[j]->cl;
		    cl2 = vec[k]->cl;
		}
	memcpy (vec_tmp, vec, vec_n * sizeof (struct VEC *));
	j = 0;
	while (vec[j]->cl != cl1 && vec[j]->cl != cl2)
	    j++;
	p = j;
	for (k = p; k < vec_n; k++)
	    if (vec_tmp[k]->cl == cl1 || vec_tmp[k]->cl == cl2) {
		vec[j] = vec_tmp[k];
		vec[j++]->cl = cl1;
	    }
	for (k = p; k < vec_n; k++)
	    if (vec_tmp [k]->cl != cl1 && vec_tmp[k]->cl != cl2)
		vec[j++] = vec_tmp[k];
    }
    maxdiff = min;
    if (verbose) {
        printf ("\n");
        fflush (stdout);
    }

    sprintf (buf3, "%s.ps", baseout);
    fp = w_fopen (buf3);
    fprintf (
	fp,
	"%c!PS\n"
        "%c%cBoundingBox: %i %i 450 704\n"
        "%c%cCreator: %s, (c) P. Kleiweg 1996\n"
        "%c%cTitle: %s\n"
        "%c%cEndComments\n\n",
        '%', '%', '%',
        146 - 5 * max_len,
        695 - 12 * (vec_n - 1),
        '%', '%', programname,
        '%', '%', infile,
        '%', '%'
    );
    fprintf (
        fp,
        "/Times-Roman findfont 10 scalefont setfont\n"
        "\n"
        "/DY 12 def\n"
        "\n"
        "/Width 300 def\n"
        "\n"
        ".5 setlinewidth\n"
        "\n"
    );
    fprintf (
        fp,
        "/s {\n"
        "    dup stringwidth pop\n"
        "    neg 150 add ml sub y moveto\n"
        "    show\n"
        "    /y y DY sub def\n"
        "} bind def\n"
	"\n"
    );
    fprintf (
        fp,
        "/l {\n"
        "    /xx exch Width MAX div mul 150 add def\n"
        "    /y2 exch DY mul neg 700 add def\n"
        "    /x2 exch Width MAX div mul 150 add def\n"
        "    /y1 exch DY mul neg 700 add def\n"
        "    /x1 exch Width MAX div mul 150 add def\n"
        "    x1 y1 moveto\n"
        "    xx y1 lineto\n"
        "    xx y2 lineto\n"
        "    x2 y2 lineto\n"
        "    stroke\n"
        "} bind def\n"
        "\n"
    );
    fprintf (
        fp,
        "/ml (a) stringwidth pop def\n"
        "\n"
    );
    fprintf (
        fp,
        "/MAX %f def\n"
        "\n",
        maxdiff
    );
    fprintf (
        fp,
        "/y 700\n"
        "gsave\n"
        "  newpath\n"
        "  0 0 moveto\n"
        "  (-) false charpath\n"
        "  pathbbox\n"
        "grestore\n"
        "exch pop add exch pop\n"
        "2 div\n"
        "sub def\n"
        "\n"
    );

    for (i = 0; i < vec_n; i++)
	fprintf (
	    fp,
	    "(%s) s\n",
	    to_ps (vec [i]->name)
	);

    if (verbose) {
	printf ("Merging Clusters");
	fflush (stdout);
    }
    for (i = 0; i < vec_n; i++) {
	vec [i]->cl = i;
	cl [i].distance = 0;
	cl [i].center = i;
    }
    for (i = 1; i < vec_n; i++) {
	if (verbose) {
	    printf (".");
	    fflush (stdout);
	}
	min = MAXFLOAT;
	for (j = 0; j < vec_n; j++)
	    for (k = j + 1; k < vec_n; k++)
		if (d [vec[j]->nr][vec[k]->nr] < min && vec[j]->cl != vec[k]->cl) {
		    min = d [vec[j]->nr][vec[k]->nr];
		    cl1 = vec[j]->cl;
		    cl2 = vec[k]->cl;
		}
	fprintf (
	    fp,
	    "%f %f %f %f %f l\n",
	    cl [cl1].distance,
	    cl [cl1].center,
	    cl [cl2].distance,
	    cl [cl2].center,
	    min
	);
	cl [cl1].distance = min;
	cl [cl1].center = (cl[cl1].center + cl[cl2].center) / 2.0;
	for (j = 0; j < vec_n; j++)
	    if (vec[j]->cl == cl2)
		vec[j]->cl = cl1;
    }
    if (verbose) {
        printf ("\n");
        fflush (stdout);
    }

    fprintf (
        fp,
        "\n"
        "showpage\n"
    );
    fclose (fp);
}

void syntax ()
{
    fprintf (
	stderr,
        "\nMerge Clustering PostScript Generator\n"
        "\n(c) P. Kleiweg 1996\n"
        "\nSyntax: %s -i string [-o string] [-q]\n"
        "-i: invoer-bestand\n"
        "-o: naam zonder extensie voor PostScript bestand\n"
        "-q: quiet\n\n",
	programname
    );
    exit (1);
}

void setup ()
{
    int
	i;

    d = (float **) m_malloc ((vec_n) * sizeof (float *));
    for (i = 0; i < vec_n; i++)
	d [i] = (float *) m_malloc (vec_n * sizeof (float));

    vec_tmp = (struct VEC **) m_malloc (vec_n * sizeof (struct VEC *));
    cl = (struct CLUSTER *) m_malloc (vec_n * sizeof (struct CLUSTER));
}

float square_vecdiff (float *v1, float *v2)
{
    int
	i;
    float
	err,
	sum;

    sum = 0.0;
    for (i = 0; i < vec_size; i++) {
	err = v1 [i] - v2 [i];
	sum += err * err;
    }
    return sum;
}

void process_args (int argc, char *argv [])
{
    int
	index;

    if (argc == 1)
        syntax ();

    for (index = 1; index < argc; index++) {
	if (argv [index][0] == '-') {
	    switch (argv [index][1]) {
		case 'q':
		    verbose = 0;
		    break;
		case 'i':
		    infile = m_strdup (get_arg (&argc, argv , &index));
		    break;
		case 'o':
		    baseout = m_strdup (get_arg (&argc, argv , &index));
#ifdef __MSDOS__
		    fnsplit (baseout, NULL, NULL, buf1, buf2);
		    if (! buf1 [0])
			errit ("Missing file name: \"%s\"", baseout);
		    if (buf2 [0])
			errit ("No file extension allowed: \"%s\"", baseout);
#endif
		    break;
		default:
                    errit ("Invalid option '%s'", argv [index]);
	    }
	} else
            errit ("Invalid parameter \"%s\"", argv [index]);
    }
}

char *get_arg (int *argc, char *argv [], int *index)
{
    int
	i;
    i = 1;
    while (argv [*index][++i]) {
	if (! isspace (argv [*index][i]))
	    return argv [*index] + i;
    }
    if (*index == *argc - 1)
	errit ("Missing argument for \"%s\"", argv [*index]);
    return argv [++*index];
}

void *m_realloc (void *block, size_t size)
{
    void
	*p;

    p = realloc (block, size);
    if (! p) {
        free (no_mem_buffer);
	errit (Out_of_memory);
    }
    return p;
}

void *m_malloc (size_t size)
{
    void
	*p;

    p = malloc (size);
    if (! p) {
        free (no_mem_buffer);
	errit (Out_of_memory);
    }
    return p;
}

char *m_strdup (const char *s)
{
    char
	*p = strdup (s);
    if (! p) {
        free (no_mem_buffer);
	errit (Out_of_memory);
    }
    return p;
}

void errit (char const *format, ...)
{
    va_list
	list;

    fprintf (stderr, "\nError %s: ", programname);

    va_start (list, format);
    vfprintf (stderr, format, list);

    fprintf (stderr, "\n\n");

    exit (1);
}

int getline (FILE *fp, int required, char const *filename)
/* Lees een regel in
 * Plaats in buffer
 * Negeer lege regels en regels die beginnen met #
 */
{
    int
	i;

    for (;;) {
	if (fgets (buffer, BUFSIZE, fp) == NULL) {
	    if (required)
		errit ("Unexpected end of file in \"%s\"", filename);
	    else
		return 0;
	}
	input_line++;
	i = strlen (buffer);
	while (i && isspace (buffer [i - 1]))
	    buffer [--i] = '\0';
	i = 0;
	while (buffer [i] && isspace (buffer [i]))
	    i++;
	if (buffer [i] == '#')
	    continue;
	if (buffer [i]) {
	    memmove (buffer, buffer + i, strlen (buffer) + 1);
	    return 1;
	}
    }
}

struct VEC *store_vec (char *name)
{
    if (vec_n == vec_max) {
	vec_max += 256;
	vec = (struct VEC **) m_realloc (vec, vec_max * sizeof (struct VEC *));
    }

    vec [vec_n] = (struct VEC *) m_malloc (sizeof (struct VEC));
    vec [vec_n]->name = m_strdup (name);
    vec [vec_n]->nr = vec_n;

    return vec [vec_n++];
}

FILE *r_fopen (char const *filename)
{
    FILE
	*fp;

    fp = fopen (filename, "r");
    if (! fp)
	errit ("Bestand \"%s\" niet gevonden", filename);
    input_line = 0;
    return fp;
}

FILE *w_fopen (char const *filename)
{
    FILE
	*fp;

    fp = fopen (filename, "w");
    if (! fp)
	errit ("Bestand \"%s\" kan niet geopend worden", filename);
    return fp;
}

void get_programname (char const *argv0)
{
#ifdef __MSDOS__
    char
        name [MAXFILE];
    fnsplit (argv0, NULL, NULL, name, NULL);
    programname = strdup (name);
#else   /* unix */
    char
        *p;
    p = strrchr (argv0, '/');
    if (p)
        programname = strdup (p + 1);
    else
        programname = strdup (argv0);
#endif    
}

char *to_ps (char const *string)
{
    int
        i, j;

    i = j = -1;
    while (string [++i])
        switch (string [i]) {
            case '(':
            case ')':
            case '\\':
                buffer [++j] = '\\';
                /*  no break  */
    	default:
                buffer [++j] = string [i];
	}
    buffer [++j] = '\0';

    return buffer;
}
