//  CLUSVIEW.CPP

#ifndef __BORLANDC__
#ifndef __TURBOC__
#error Borland C or Turbo C compiler required
#endif
#endif

#ifndef __COMPACT__
#error Memory model COMPACT required
#endif
#include <alloc.h>
#include <bios.h>
#include <ctype.h>
#include <dir.h>
#include <graphics.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <values.h>

#define BUFSIZE 2048

struct VEC
{
    char
	*name;
    float
	*v;
    int
	cl,
        nr;
};

struct CLUSTER
{
    float
	distance,
	center;
};

struct VEC
    **vec = NULL,
    **vec_tmp = NULL;

struct CLUSTER
    *cl;

float
    **d;

int
    vec_max = 0,
    vec_n = 0,
    vec_size,
    input_line,
    in_graph = 0;

char
    *programname,
    buffer [BUFSIZE],
    buf1 [BUFSIZE],
    *no_mem_buffer,
    Out_of_memory [] = "Out of memory";

void
    get_programname (char const *argv0),
    syntax (void),
    setup (void),
    read_file (char const *filename),
    show_clusters (void),
    errit (char const *format, ...),
    *m_malloc (size_t size),
    *m_realloc (void *block, size_t size);
char
    *m_strdup (const char *s);
int
    getline (FILE *fp, int required, char const *filename);
float
    square_vecdiff (float *v1, float *v2);
FILE
    *r_fopen (char const *filename);
struct VEC
    *store_vec (char *name);

int main (int argc, char *argv [])
{
    no_mem_buffer = (char *)malloc (1024);

    get_programname (argv [0]);

    if (argc != 2)
	syntax ();

    read_file (argv [1]);

    setup ();

    show_clusters ();

    return 0;
}

void read_file (char const *infile)
{
    FILE
	*fp;
    struct VEC
	*v;
    int
	i;

    fp = r_fopen (infile);
    getline (fp, 1, infile);
    if (sscanf (buffer, "%i", &vec_size) != 1)
	errit (
	    "file \"%s\", line %i\nVectorsize expected",
	    infile,
	    input_line
	);
    while (getline (fp, 0, NULL)) {
	if (sscanf (buffer, " %s", buf1) != 1)
	    errit ("file \"%s\", line %i", infile, input_line);
	v = store_vec (buf1);
	v->v = (float *) m_malloc (vec_size * sizeof (float));
	for (i = 0; i < vec_size; i++) {
	    getline (fp, 1, infile);
	    if (sscanf (buffer, "%f", &(v->v[i])) != 1)
		errit (
		    "file \"%s\", line %i\n"
		    "Missing value for vector \'%s\'",
		    infile,
		    input_line,
		    v->name
		);
	}
    }
    fclose (fp);
}

void show_clusters ()
{
    int
	i,
	j,
	k,
	p,
	max_len,
	cl1,
	cl2,
	gdriver = DETECT,
	gmode,
	errorcode;
    float
	min,
	x1,
	x2,
	xx,
	xd,
	y1,
	y2,
        yd,
	xmin,
	ymin;

    max_len = 0;
    for (i = 0; i < vec_n; i++)
	if ((j = strlen (vec [i]->name)) > max_len)
	    max_len = j;

    printf ("Calculating Sums of Square Differences");
    for (i = 0; i < vec_n; i++) {
	printf (".");
	for (j = i; j < vec_n; j++)
            d [i][j] = d [j][i] = square_vecdiff (vec [i]->v, vec [j]->v);
    }
    printf ("\n");

    printf ("Sorting");
    for (i = 0; i < vec_n; i++)
	vec [i]->cl = i;
    for (i = 1; i < vec_n; i++) {
	printf (".");
	min = MAXFLOAT;
	for (j = 0; j < vec_n; j++)
	    for (k = j + 1; k < vec_n; k++)
		if (d [vec[j]->nr][vec[k]->nr] < min && vec[j]->cl != vec[k]->cl) {
		    min = d [vec[j]->nr][vec[k]->nr];
		    cl1 = vec[j]->cl;
		    cl2 = vec[k]->cl;
		}
	memcpy (vec_tmp, vec, vec_n * sizeof (struct VEC *));
	j = 0;
	while (vec[j]->cl != cl1 && vec[j]->cl != cl2)
	    j++;
	p = j;
	for (k = p; k < vec_n; k++)
	    if (vec_tmp[k]->cl == cl1 || vec_tmp[k]->cl == cl2) {
		vec[j] = vec_tmp[k];
		vec[j++]->cl = cl1;
	    }
	for (k = p; k < vec_n; k++)
	    if (vec_tmp [k]->cl != cl1 && vec_tmp[k]->cl != cl2)
		vec[j++] = vec_tmp[k];
    }
    printf ("\n");

    registerbgidriver (EGAVGA_driver);
    initgraph (&gdriver, &gmode, getenv ("BGI"));
    errorcode = graphresult();
    if (errorcode != grOk) {
	errit (
	    "%s\n\n"
	    "Environment variable BGI can be used to point to\n"
	    "directory where graphic driver files are to be found",
	    grapherrormsg (errorcode)
	);
    }
    in_graph = 1;
    settextjustify (RIGHT_TEXT, CENTER_TEXT);
    // setlinestyle (SOLID_LINE, 0, 1);

    sprintf (buffer, "%*s", max_len, "");
    xmin = textwidth (buffer) + 4;
    xd = ((float)(getmaxx () - xmin)) / min;
    yd = getmaxy () / ((vec_n > 10) ? vec_n : 10);
    ymin = yd / 2;

    for (i = 0; i < vec_n; i++)
	outtextxy (xmin - 4, ymin + i * yd, vec [i]->name);

    for (i = 0; i < vec_n; i++) {
	vec [i]->cl = i;
	cl [i].distance = 0;
	cl [i].center = i;
    }
    for (i = 1; i < vec_n; i++) {
	min = MAXFLOAT;
	for (j = 0; j < vec_n; j++)
	    for (k = j + 1; k < vec_n; k++)
		if (d [vec[j]->nr][vec[k]->nr] < min && vec[j]->cl != vec[k]->cl) {
		    min = d [vec[j]->nr][vec[k]->nr];
		    cl1 = vec[j]->cl;
		    cl2 = vec[k]->cl;
		}
	x1 = xmin + xd * cl [cl1].distance;
	x2 = xmin + xd * cl [cl2].distance;
	xx = xmin + xd * min;
	y1 = ymin + yd * cl [cl1].center;
	y2 = ymin + yd * cl [cl2].center;
	line (x1, y1, xx, y1);
	line (xx, y1, xx, y2);
	line (xx, y2, x2, y2);
	cl [cl1].distance = min;
	cl [cl1].center = (cl[cl1].center + cl[cl2].center) / 2.0;
	for (j = 0; j < vec_n; j++)
	    if (vec[j]->cl == cl2)
		vec[j]->cl = cl1;
    }

    _bios_keybrd (_KEYBRD_READ);
    closegraph ();
    in_graph = 0;
}

void syntax ()
{
    fprintf (
	stderr,
	"\nMerge Clustering Viewer\n"
        "\n(c) P. Kleiweg 1996\n"
        "\nSyntax: %s file\n",
	programname
    );
    exit (1);
}

void setup ()
{
    int
	i;

    d = (float **) m_malloc ((vec_n) * sizeof (float *));
    for (i = 0; i < vec_n; i++)
	d [i] = (float *) m_malloc (vec_n * sizeof (float));

    vec_tmp = (struct VEC **) m_malloc (vec_n * sizeof (struct VEC *));
    cl = (struct CLUSTER *) m_malloc (vec_n * sizeof (struct CLUSTER));
}

float square_vecdiff (float *v1, float *v2)
{
    int
	i;
    float
	err,
        sum;

    sum = 0.0;
    for (i = 0; i < vec_size; i++) {
	err = v1 [i] - v2 [i];
        sum += err * err;
    }
    return sum;
}

void *m_realloc (void *block, size_t size)
{
    void
	*p;

    p = realloc (block, size);
    if (! p) {
	free (no_mem_buffer);
	errit (Out_of_memory);
    }
    return p;
}

void *m_malloc (size_t size)
{
    void
	*p;

    p = malloc (size);
    if (! p) {
        free (no_mem_buffer);
	errit (Out_of_memory);
    }
    return p;
}

char *m_strdup (const char *s)
{
    char
	*p = strdup (s);
    if (! p) {
        free (no_mem_buffer);
	errit (Out_of_memory);
    }
    return p;
}

void errit (char const *format, ...)
{
    va_list
	list;

    if (in_graph) {
	closegraph ();
	in_graph = 0;
    }

    fprintf (stderr, "\nError %s: ", programname);

    va_start (list, format);
    vfprintf (stderr, format, list);

    fprintf (stderr, "\n\n");

    exit (1);
}

int getline (FILE *fp, int required, char const *filename)
/* Lees een regel in
 * Plaats in buffer
 * Negeer lege regels en regels die beginnen met #
 */
{
    int
	i;

    for (;;) {
	if (fgets (buffer, BUFSIZE, fp) == NULL) {
	    if (required)
		errit ("Unexpected end of file in \"%s\"", filename);
	    else
		return 0;
	}
	input_line++;
	i = strlen (buffer);
	while (i && isspace (buffer [i - 1]))
	    buffer [--i] = '\0';
	i = 0;
	while (buffer [i] && isspace (buffer [i]))
	    i++;
	if (buffer [i] == '#')
	    continue;
	if (buffer [i]) {
	    memmove (buffer, buffer + i, strlen (buffer) + 1);
	    return 1;
	}
    }
}

struct VEC *store_vec (char *name)
{
    if (vec_n == vec_max) {
	vec_max += 256;
	vec = (struct VEC **) m_realloc (vec, vec_max * sizeof (struct VEC *));
    }

    vec [vec_n] = (struct VEC *) m_malloc (sizeof (struct VEC));
    vec [vec_n]->name = m_strdup (name);
    vec [vec_n]->nr = vec_n;

    return vec [vec_n++];
}

FILE *r_fopen (char const *filename)
{
    FILE
	*fp;

    fp = fopen (filename, "r");
    if (! fp)
	errit ("Bestand \"%s\" niet gevonden", filename);
    input_line = 0;
    return fp;
}

void get_programname (char const *argv0)
{
    char
	name [MAXFILE];
    fnsplit (argv0, NULL, NULL, name, NULL);
    programname = strdup (name);
}
